#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 30 11:48:05 2019

@author: zandieh
"""
import numpy as np
import torch
import torch.fft
from torch import linalg as LA

import math
import quadprog



def quadprog_solve_qp(P, q, G=None, h=None, A=None, b=None):
    qp_G = .5 * (P + P.T +0.00001*np.eye(P.shape[0]))   # make sure P is symmetric
    qp_a = -q
    if A is not None:
        qp_C = -np.vstack([A, G]).T
        qp_b = -np.hstack([b, h])
        meq = A.shape[0]
    else:  # no equality constraint
        qp_C = -G.T
        qp_b = -h
        meq = 0
    return quadprog.solve_qp(qp_G, qp_a, qp_C, qp_b, meq)[0]

def NTK(L,q):
    n=15*L+5*q
    Y = np.zeros((201+n,L+1))
    Y[:,0] = np.sort(np.concatenate((np.linspace(-1.0, 1.0, num=201), np.cos((2*np.arange(n)+1)*np.pi / (4*n))), axis=0))

    m_ = Y.shape[0]

    for i in range(L):
        Y[:,i+1] = (np.sqrt(1-Y[:,i]**2) + Y[:,i]*(np.pi - np.arccos(Y[:,i])))/np.pi
    
    y = np.zeros(m_)
    for i in range(L+1):
        z = Y[:,i]
        for j in range(i,L):
            z = z*(np.pi - np.arccos(Y[:,j]))/np.pi
        y = y + z
    
    Z = np.zeros((m_,q+1))
    Z[:,0] = np.ones(m_)
    for i in range(q):
        Z[:,i+1] = Z[:,i] * Y[:,0]
    
    
    weight_ = np.linspace(0.0, 1.0, num=m_) + 2/L
    w = y * weight_
    U = Z.T * weight_

    beta_ = quadprog_solve_qp(np.dot(U, U.T), -np.dot(U,w) , np.concatenate((Z[0:m_-1,:]-Z[1:m_,:], -np.eye(q+1)),axis=0), np.zeros(q+m_))
    beta_[beta_ < 0.00001] = 0
    
    return beta_



class RNSampler:
    
    def __init__(self, m):
            
        self.m = m
        
        self.S = [0 for i in range(m)]
        self.W = torch.zeros(m, dtype=torch.float64)
        
    
    def Gen_Samples(self, s, q, X):
        
        self.n = X.shape[0]
        self.d = X.shape[1]
        self.s = s
        
        J = np.random.choice(self.n, self.m)
        
        Norms = LA.norm(X[J,:], dim=1)**2
        
        y = torch.ones(q+1, self.m)
        for i in range(1,q+1):
            y[i,:] = y[i-1,:] * (2*Norms/s)/i
        y = y/torch.sum(y, dim=0)
        
        y = y.data.numpy()
        
        
        Exp = np.exp(-(LA.norm(X, dim=1)**2)/s)
        
        for l, j in enumerate(J):
            
            w = np.random.choice(q+1, 1, p=y[:,l])[0]
            
            I = np.random.choice(self.d, w, p = (X[j,:]**2)/Norms[l] )
            
            beta_ = Exp
            
            for a in range(w):
                beta_ = beta_ * X[:, I[a]]
                
            
            self.W[l] = np.sqrt(self.n/self.m) / LA.norm( beta_ )
            
            self.S[l] = torch.from_numpy( np.concatenate((np.array([w]), I)) )


    def Apply_Sampler(self, X):
        
        n = X.shape[0]
                
        Z = (torch.ones(self.m, n) * np.exp(-(LA.norm(X, dim=1)**2)/self.s)).T * self.W
        
        for i in range(self.m):
            
            for a in range(self.S[i][0]):
                
                Z[:,i] = Z[:,i] * X[:, self.S[i][1+a]]
        
        return Z
    
    
class NTK_RNSampler:
    
    def __init__(self, m):
            
        self.m = m
        
        self.S = [0 for i in range(m)]
        self.W = torch.zeros(m, dtype=torch.float64)
        
    
    def Gen_Samples(self, L, q, X):
        
        self.n = X.shape[0]
        self.d = X.shape[1]
        self.L = L
        
        J = np.random.choice(self.n, self.m)
        
        self.NTK_coeff = NTK(L, q)
        
        y = self.NTK_coeff
        y = y/np.sum(y)
        
        w = np.random.choice(q+1, self.m, p=y)
        
        norm_X = LA.norm(X, dim=1)
        Normalizer = torch.where(norm_X>0, norm_X, 1.0)
        Xnormalized = ((X.T / Normalizer).T)
        
        for l, j in enumerate(J):
            
            I = np.random.choice(self.d, w[l], p = Xnormalized[j,:]**2 )
            
            beta_ = norm_X
            
            for a in range(w[l]):
                beta_ = beta_ * Xnormalized[:, I[a]]
                
            
            self.W[l] = np.sqrt(2*self.n/self.m) / LA.norm( beta_ )
            
            self.S[l] = torch.from_numpy( np.concatenate((np.array([w[l]]), I)) )


    def Apply_Sampler(self, X):
        
        n = X.shape[0]
        
        norm_X = LA.norm(X, dim=1)
        Normalizer = torch.where(norm_X>0, norm_X, 1.0)
        Xnormalized = ((X.T / Normalizer).T)

        Z = (torch.ones(self.m, n) * norm_X).T * self.W
        
        for i in range(self.m):
            
            for a in range(self.S[i][0]):
                
                Z[:,i] = Z[:,i] * Xnormalized[:, self.S[i][1+a]]
        
        return Z